#!/usr/bin/perl
#
# lt24merge.pl - Merge LED Toy firmware with images
#
# Written 2005, 2006 by Werner Almesberger
# Copyright 2005, 2006 by Werner Almesberger
#

#
# PIC memory
#
$MAX_BYTES = (4096-0x100)*2;

#
# Timing
#
$T = 1000000/256/256;	# timer cycle

#
# n-th diode to port and phase:
#
# Value Port Phase
#  0- 7  B     1
#  8-15  A     1
# 16-23  B     2
# 24-31  A     2
#

#
# The 12 pixels in a group, which are spread out over 16 ports, are packed
# into a 14 bit value, together with the phase select bits.
#
# LED     Port  Encoding  Comment      
#  0,  1  RB1   1+8      
#  2,  3  RB3   3+8  
#  4,  5  RB2   2+8
#  6,  7  RB6   5         instead of unused RA5
#  8,  9  RB4   4+8
# 10, 11  RB7   5+8       instead of unused RB5       
# 12, 13  RA6   6   
# 14, 15  RA7   7
# 16, 17  RA0   0
# 18, 19  RA1   1
# 20, 21  RA2   2                
# 22, 23  RA3   3                  
#

#
# The original "A" map
#

@map =
  ( 1+ 8, 1+24, 3+ 8, 3+24, 2+ 8, 2+24, 5,    5+16,
    4+ 8, 4+24, 5+ 8, 5+24, 6,    6+16, 7,    7+16,
    0,    0+16, 1,    1+16, 2,    2+16, 3,    3+16 );
$xor = 0;


sub swap
{
    local ($a,$b) = @_;

    ($map[$a],$map[$a+1],$map[$b],$map[$b+1]) =
      ($map[$b],$map[$b+1],$map[$a],$map[$a+1]);
}


#
# RA4 selects phase 1, RB0 selects phase 2
#

$PHASE_SELECT = (1 << 4) | (1 << (0+24));


sub convert
{
    local $width = shift(@_);
    local ($x,$y);
    local @res = ();

    for ($x = 0; $x != $width; $x++) {
	$pat = 0;
	for ($y = 0; $y != 24; $y++) {
	    $pix = $x+$y*(($width+7) & ~7);
	    $pat |= 1 << $map[$y] if $_[$pix >> 3] & (1 << ($pix & 7));
	}
	push(@res,$pat);
    }
    return @res;
}


sub load_xbm
{
    local ($s,$x,$bytes,@res);

    open(FILE,$_[0]) || die "$_[0]: $!";
    $s = join("",<FILE>);
    close FILE;
    $s =~ /#define\s\S+_width\s(\d+)\s+#define\s\S+_height\s24\s+/ ||
      die "$_[0]: not a valid X bitmap file";
    die "image width $1 out of range [1, 255]" if $1 < 1 || $1 > 255;
    $x = $1;
    $bytes = 3*(($x+7) & ~7)-1;
    $' =~ /^{0}static\s+(unsigned\s+)*char\s+\S+_bits\[\]\s=\s\{\s+
{0}(0x[0-9a-f]{2}(,\s*0x[0-9a-f]{2}){$bytes})\};\s*$/ ||
      die "$_[0]: not a valid X bitmap file";
    return ($x,map { hex $_; } split(/,\s*/,$2));
}


sub load_img
{
    local $spec = $_[0];
    local ($x,$i,$res);
    local @data;
    local $flyback = 0;

    if ($spec =~ m|\s*//\s*(\d+)\s*$|) {
	$flyback = $1;
	$spec = $`;
    }
    while (length $spec) {
	local ($op,$not,$cx,@cd);

        $spec =~ /^\s*([|&^])\s*|/;
	($op,$spec) = ($1,$');
	if ($op && !defined $x) {
	    die "image spec must not begin with an operator \"$&\"";
	}
	if (!$op && defined $x) {
	    die "file names not joined by an operator \"$&\"";
	}
	$spec =~ /^\s*(~?)\s*([^&|]+)([|&^]?.*|$)/ ||
	  die "garbled image spec \"$spec\"";
	$not = $1 eq "~";
	$spec = $3;
	($cx,@cd) = &load_xbm($2);
	map { $_ = ~$_ } @cd if $not;
	if (!$op) {
	    $x = $cx;
	    @data = @cd;
	}
	else {
	    local $i;

	    die "cannot combine images with different sizes ($cx != $x)"
	      if $x != $cx;
	    for ($i = 0; $i != @data; $i++) {
		$data[$i] = $op eq "&" ? $data[$i] & $cd[$i] :
		  $op eq "|" ? $data[$i] | $cd[$i] : $data[$i] ^ $cd[$i];
	    }
	}
    }
    @data = &convert($x,@data);
    if ($x == 24) {	# for historical reasons, we only trim 24x24 images
	local ($left,$right);

	for ($left = 0; $left != $x; $left++) {
	    last if $data[$left];
	}
	for ($i = $left; $i != $x; $i++) {
	    $right = $i if $data[$i];
	}
	for ($i = 0; $i != $left; $i++) {
	    shift(@data);
	}
	$x = $right-$left+1;
    }
    print STDERR "warning: flyback ($flyback) is less than image width ($x)\n"
      if $flyback && $flyback < $x;
    die "flyback value $flyback is too large (> 1008)" if $flyback > 1008;
    $res = pack("CC",$x,($flyback+15) >> 4);
    for ($i = 0; $i != $x; $i++) {
	$res .= pack("V",($data[$i] | $PHASE_SELECT) ^ $xor);
    }
    return $obj{$res} if defined $obj{$res};
    $obj{$res} = pack("v",@ARGV+1+(length($data)+length($bin))/2+0x100);
    $data .= $res;
    return $obj{$res};
}


sub load_seq
{
    local ($s,$res,$end,$retval);

    open(FILE,$_[0]) || die "$_[0]: $!";
    $s = join("",<FILE>);
    close FILE;
    $s =~ s/\s*#[^\n]*\n+/\n/g;
    $s =~ /title\s+([^\n]*?\S)\s*?\n/ || die "$_[0]: no title object";
    $s = $';
    $res = pack("v",0x200).&load_img($1);
    $end = 0;
    while (length $s) {
	local $flags;

	die "$_[0]: data beyond the end" if $end;
	$s =~ /^image\s+([^\n]*?\S)\s*rotate\s+([0123])\s*
{0}timeout\s+(\d+(\.\d*)?)\s*(loop|end)?\s*/ ||
	  die "$_[0]: invalid object entry";
	$s = $';
	$end = $5 ne "";
	$flags = 0;
	$flags |= 1 if $5 eq "end";
	$flags |= 2 if $5 eq "loop";
	$res .= pack("CC",$2,$flags);
	$res .= pack("v",$3*$T);
	$res .= &load_img($1);
    }
    die "$_[0]: no end marker" if !$end;
    $retval = pack("v",@ARGV+1+(length($data)+length($bin))/2+0x100);
    $data .= $res;
    return $retval;
}


$fw = shift(@ARGV);
if (!defined $fw) {
    print STDERR "usage: $ARGV[0] firmware.bin [image.xbm ...]\n";
    exit(1);
}

if ($fw =~ /-LB24A\.bin$/) {
    # Nothing to do
}
elsif ($fw =~ /-LB24(C|V[01])\.bin$/) {
    #
    # V0 map is like A map, but with RB4 and RB6 swapped
    #
    &swap(6,8);
}
else {
    #
    # V2 map is like V0, but with RB2 and RB3 swapped and all polarities
    # reversed. V2, V3, and V4 have identical firmware.
    #
    print STDERR
      "warning: unknown firmware version, assuming V4 compatibility\n"
      if $fw !~ /-LB24V[234]\.bin$/;
    &swap(6,8);
    &swap(2,4);
    $xor = 0x3fff3fff;
}


open(FILE,$fw) || die "$fw: $!";
while (<FILE>) {
    $bin .= $_;
}
close FILE;

$table = "";
$data = "";
for (@ARGV) {
    if ($_ =~ m|\.xbm(\s*//\s*\d+)?$|) {
	$table .= &load_img($_);
    }
    elsif ($_ =~ /\.seq$/) {
	$table .= &load_seq($_);
    }
    else {
	die "$_ is not a bitmap (.xbm) or sequence (.seq) file";
    }
}

($len_fw,$len_data) = (length $bin,2+length $data);
$bin .= $table.pack("v",0).$data;

$len = length($bin);
if ($len > $MAX_BYTES) {
    print STDERR "output is too big ($len > $MAX_BYTES)\n";
    exit(1);
}

print $bin;

print STDERR $len/2,"/",$MAX_BYTES/2," words (",int(100*$len/$MAX_BYTES+0.5),
  "%): firmware ",$len_fw/2," (",int(100*$len_fw/$MAX_BYTES+0.5),
  "%), data ",$len_data/2," (",int(100*$len_data/$MAX_BYTES+0.5),"%)\n";
