;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Firmware upgrade monitor, with XMODEM receiver
;
; Written 2005, 2006 by Werner Almesberger
; Copyright 2005, 2006 by Werner Almesberger
;
; The XMODEM protocol is described in chapter 7 of "XMODEM/YMODEM PROTOCOL
; REFERENCE" by Chuck Forsberg,
; http://timeline.textfiles.com/1988/10/14/1/FILES/ymodem.txt
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; XMODEM message types

SOH		=	0x01
EOT		=	0x04
ACK		=	0x06
NAK		=	0x15
CAN		=	0x18

; Protocol constants

XM_RETRIES	=	  4	; the number of times we retry a block, 4.
				; The protocol reference suggests 10, but since
				; we usually have an excellent line, this just
				; makes failure recovery very slow.
XM_TO1ST	=	 48	; timeout for first byte in packet, 3 s
				; This is shorter than the 10 s timeout
				; suggested in the protocol reference, to make
				; downloads start more rapidly.
XM_TOREST	=	 16	; timeout for all other bytes in packet, 1 s
XM_TOPURGE	=	 10	; timeout for clearing the line, 0.6 s
				; This is shorter than the 1 s timeout
				; suggested in the protocol reference, to allow
				; for senders timing out after 1 s. (Just in
				; case ...)
XM_BLKLEN	=	128	; blocks are 128 bytes each

	cblock	last_global+1

xmto			; timeout (ticks at 1/16 Hz)
xmnext			; sequence of next packet we expect
xmmore			; number of additional bytes we expect
xmchk			; checksum
xmretry			; number of retries left
xmtmp			; scratch register for low-level functions

xmbuf			; receive buffer starts here
	endc


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Macros
;

; Reset the buffer

resetbuf macro
	movlw	xmbuf
	movwf	FSR
	endm


; Add byte in W to the buffer

putbuf:	macro
	movwf	INDF
	call	nextbuf
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Enter the XMODEM receiver if anything is connected to the serial port
;

; Enter the monitor if online

pollmon:
	call	online		; return if we aren't online
	btfsc	STATUS,Z
	return

	; fall through to "monloop"

; The monitor main loop

monloop:
	mnldon			; show that we're listening

	; Make T0 tick at 16 Hz (without using interrupts)

	bcf	INTCON,GIE	; no interrupts
	bsf	STATUS,RP0	; run at OSC/4
	bcf	OPTION_REG,T0CS
	bcf	OPTION_REG,PSA	; the prescaler is ours
	; prescaler defaults to 256 on reset

	bcf	STATUS,RP0

	; Set up the UART

	movlw	(1 << SPEN) | (1 << CREN) ; enable serial port and receiver
	movwf	RCSTA
	bsf	STATUS,RP0
	movlw	12		; 19200 bps
	movwf	SPBRG
	movlw	(1 << TXEN) | (1 << BRGH) ; 19200 bps, enable transmit
	movwf	TXSTA
	bcf	STATUS,RP0
	
	call	purge		; skip any line noise

	movlw	1		; try to receive the first block
	movwf	xmnext
	call	rxblk
	btfsc	STATUS,Z	; EOF -> shut down
	goto	reset

	bsf	STATUS,RP1	; start programming at 0x100
	movlw	1
	movwf	EEADRH
	clrf	EEADR
	bcf	STATUS,RP1

	; fall through to "download"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Download a firmware image
;

download:
	resetbuf		; to the buffer start

	movlw	64		; 128 data bytes left
	movwf	tmp

pgmlp:

	bsf	STATUS,RP1

	movlw	31		; every 32th word, erase
	andwf	EEADR,W
	btfss	STATUS,Z
	goto	pgmone

	bsf	STATUS,RP0	; prepare for erasing
	movlw	(1 << EEPGD) | (1 << WREN) | (1 << FREE)
	movwf	EECON1

	movlw	0x55		; magic sequence
	movwf	EECON2
	movlw	0xaa
	movwf	EECON2
	
	bsf	EECON1,WR	; do the erase
	nop
	nop

pgmone:	; STATUS:RP1 is set, STATUS:RP0 may be set

	bsf	STATUS,RP0	; prepare to write
	movlw	(1 << EEPGD) | (1 << WREN)	
	movwf	EECON1

	bcf	STATUS,RP0	; get the next byte
	bcf	STATUS,RP1
	call	getbuf
	
	bsf	STATUS,RP1	; store it
	movwf	EEDATA
	bcf	STATUS,RP1

	call	getbuf		; get another byte

	bsf	STATUS,RP1	; store it too
	movwf	EEDATH

	bsf	STATUS,RP0
	movlw	0x55		; magic sequence
	movwf	EECON2
	movlw	0xaa
	movwf	EECON2

	bsf	EECON1,WR	; do the write
	nop
	nop

	bcf	STATUS,RP0	; move write pointer
	incf	EEADR,F
	movf	EEADR,F
	btfsc	STATUS,Z
	incf	EEADRH,F

	bcf	STATUS,RP1	; next byte
	decfsz	tmp,F
	goto	pgmlp

	movlw	ACK		; send an ACK
	movwf	TXREG

	call	rxblk		; receive the next block
	btfss	STATUS,Z
	goto	download	; more (not EOF) -> continue

	; fall through to reset

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Force a watchdog reset
;

reset:
	; interrupts are already off
	; the prescaler is already assigned to T0

	bsf	STATUS,RP1	; enable the watchdog timer, period 1:32
	movlw	1 << SWDTEN
	movwf	WDTCON

waitdog:
	goto	waitdog		; wait for the watchdog to bite


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Online check
;

; Reset if offline

chkline:
	call	online		; return if online
	btfss	STATUS,Z
	return

	goto	reset		; force a reset


; Set Z if offline. We're offline if RX stays low for at least one frame.

online:
	movlw	255		; wait for about 1 ms to let line settle
	movwf	tmp
onxlp:
	goto	onxdg		; 2
onxdg:
	decfsz	tmp,F		; 4

	movlw	(521+4)/5+1	; probe for 0.521 ms (10 bits at 19.2 kbps)
	movwf	tmp
onllp:
	btfsc	RX		; is RX high ?		; 2
	goto	onlok

	decfsz	tmp,F		; loop back 		; 3
	goto	onllp					; 5

	bsf	STATUS,Z	; return with Z set
	return

onlok:
	bcf	STATUS,Z	; return with Z clear
	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Receive a block, with all timeouts, retries, etc.
;

; Returns with Z set on EOF

rxblk:
	movlw	XM_RETRIES	; ten retries
	movwf	xmretry

rxdoblk:
	mnldon			; show that we're listening

	movlw	XM_TO1ST	; timeout for the first byte
	movwf	xmto

rx1st:
	call	rxbyte		; get first byte
	btfsc	STATUS,Z	; timeout or error ?
	goto	txnak		; yes -> send a NACK

	xorlw	EOT		; EOT  -> EOF
	btfsc	STATUS,Z
	goto	rxeot

	xorlw	SOH ^ EOT	; SOH ?
	btfss	STATUS,Z	; no -> try again
	goto	rx1st

	call	rxbrest		; get the next byte (block number)
	btfsc	STATUS,Z	; timeout or error -> NAK
	goto	txnak

	movwf	xmtmp		; keep the block number

	call	rxbrest		; get the next byte (inverted block number)
	btfsc	STATUS,Z	; timeout or error -> NAK
	goto	txnak

	xorwf	xmtmp,W		; do they match ?
	addlw	1
	btfss	STATUS,Z
	goto	txnak		; no -> NAK

	movf	xmtmp,W		; is this the next block ?
	subwf	xmnext,W
	btfsc	STATUS,Z
	goto	rxthis		; yes -> receive it

	addlw	-1		; repetition of the previous one ?
	btfss	STATUS,Z
	goto	txnak		; no -> NAK

	; Known bug: we'll silently ACK if first block has number 0

	call	purge		; get rid of the current block

	movlw	ACK		; send an ACK
	movwf	TXREG

	goto	rxblk		; wait for more

rxthis:
	resetbuf		; start the buffer

	movlw	XM_BLKLEN	; receive 128 bytes
	movwf	xmmore
	clrf	xmchk		; clear the checksum

rxbuf:
	call	rxbrest		; get the next data byte
	btfsc	STATUS,Z	; timeout or error -> NAK
	goto	txnak

	addwf	xmchk,F		; add the bute to the checksum ...
	putbuf			; ... and put it into the buffer

	decfsz	xmmore,F	; next byte
	goto	rxbuf

	call	rxbrest		; get the checksum
	btfsc	STATUS,Z	; timeout or error -> NAK
	goto	txnak

	xorwf	xmchk,W		; checksum error -> NAK
	btfss	STATUS,Z
	goto	txnak

	incf	xmnext,F	; next block

	bcf	STATUS,Z	; return with Z clear
	return

rxeot:
	movlw	ACK		; acknowledge the EOT
	movwf	TXREG

	bsf	STATUS,RP0	; wait until the shift register is empty
eotwsr:	
	btfss	TXSTA,TRMT
	goto	eotwsr
	bcf	STATUS,RP0

	bsf	STATUS,Z	; return with Z set
	return

txnak:
	mnldoff			; briefly turn LED off to indicate trouble

	call	chkline		; exit if offline

	decfsz	xmretry,F	; too many retries ?
	goto	txdonak
	goto	monloop		; give up (note: we'll return to the main code
				; through a reset, so this open recursion is
				; fine)

txdonak:
	call	purge		; wait for the line to become quiet

	movlw	NAK		; send a NAK
	movwf	TXREG

	goto	rxdoblk		; try again

purge:
	movlw	XM_TOPURGE	; timeout for purging
	movwf	xmto

	call	rxbyte		; see if anything is coming
	btfss	STATUS,Z	; if so, wait some more
	goto	purge

	return			; line is nice and quiet now

rxbrest:
	movlw	XM_TOREST	; timeout for non-first bytes
	movwf	xmto

	; fall through to "rxbyte"


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Receive one byte, with timeouts
;
; Receives one byte in W. xmto is decremented while we wait. If it reaches
; zero, rxbyte returns, with an arbitrary value in W, and Z set. Otherwise, Z
; is cleared. Likewise, if there was any error, Z is set.
;

rxbyte:
	; Restart the timer

	movlw	-T0_16HZ	; 16 Hz
	movwf	TMR0

	bcf	INTCON,TMR0IF	; clear any pending timer interrupts

rxwlp:
	btfsc	RCSTA,OERR	; overrun
	goto	rxovr

	btfsc	RCSTA,FERR	; framing error
	goto	rxfrm

	btfsc	PIR1,RCIF	; got data to receive
	goto	rxgot

	btfss	INTCON,TMR0IF	; timer interval ?
	goto	rxwlp		; no -> keep on waiting

	decfsz	xmto,F		; timeout ?
	goto	rxbyte

	goto rxerr		; return with Z set

rxgot:
	movf	RCREG,W		; get the received byte
	bcf	STATUS,Z	; clear Z
	return			; done

rxovr:
	bcf	RCSTA,CREN	; reset the receiver
	bsf	RCSTA,CREN

	goto	rxerr		; return with Z set

rxfrm:
	movf	RCREG,W		; get the garbled byte

rxerr:
	bsf	STATUS,Z	; return with Z set
	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Buffer management (continued)
;

; Read one byte from the buffer into W

getbuf:
	movf	INDF,W
gb_safew:
	movwf	xmtmp
	call	nextbuf
	movf	xmtmp,W
	return


; Move to the next buffer location

nextbuf:
	incf	FSR,F
	movlw	0x80		; skip over 0x80...0x9f
	xorwf	FSR,W
	btfss	STATUS,Z
	return
	movlw	0x20
	addwf	FSR,F
	return
